@@ -4,7 +4,7 @@
  */
 
 /*
- * Copyright (C) 2013-2019 Oracle Corporation
+ * Copyright (C) 2013-2020 Oracle Corporation
  * This file is based on ast_mode.c
  * Copyright 2012 Red Hat Inc.
  * Parts based on xf86-video-ast
@@ -39,11 +39,11 @@
 #include "vbox_drv.h"
 #include <linux/export.h>
 #include <drm/drm_crtc_helper.h>
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0) || defined(RHEL_72)
-#include <drm/drm_plane_helper.h>
+#if RTLNX_VER_MIN(3,18,0) || RTLNX_RHEL_MAJ_PREREQ(7,2)
+# include <drm/drm_plane_helper.h>
 #endif
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0) || defined(RHEL_81)
-#include <drm/drm_probe_helper.h>
+#if RTLNX_VER_MIN(5,1,0) || RTLNX_RHEL_MAJ_PREREQ(8,1)
+# include <drm/drm_probe_helper.h>
 #endif
 
 #include "VBoxVideo.h"
@@ -63,18 +63,16 @@ static void vbox_do_modeset(struct drm_crtc *crtc,
 	struct vbox_crtc *vbox_crtc = to_vbox_crtc(crtc);
 	struct vbox_private *vbox;
 	int width, height, bpp, pitch;
-	unsigned int crtc_id;
 	u16 flags;
 	s32 x_offset, y_offset;
 
 	vbox = crtc->dev->dev_private;
 	width = mode->hdisplay ? mode->hdisplay : 640;
 	height = mode->vdisplay ? mode->vdisplay : 480;
-	crtc_id = vbox_crtc->crtc_id;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0) || defined(RHEL_75)
+#if RTLNX_VER_MIN(4,11,0) || RTLNX_RHEL_MAJ_PREREQ(7,5)
 	bpp = crtc->enabled ? CRTC_FB(crtc)->format->cpp[0] * 8 : 32;
 	pitch = crtc->enabled ? CRTC_FB(crtc)->pitches[0] : width * bpp / 8;
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 3, 0)
+#elif RTLNX_VER_MIN(3,3,0)
 	bpp = crtc->enabled ? CRTC_FB(crtc)->bits_per_pixel : 32;
 	pitch = crtc->enabled ? CRTC_FB(crtc)->pitches[0] : width * bpp / 8;
 #else
@@ -95,7 +93,7 @@ static void vbox_do_modeset(struct drm_crtc *crtc,
 	    vbox_crtc->fb_offset % (bpp / 8) == 0)
 		VBoxVideoSetModeRegisters(
 			width, height, pitch * 8 / bpp,
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0) || defined(RHEL_75)
+#if RTLNX_VER_MIN(4,11,0) || RTLNX_RHEL_MAJ_PREREQ(7,5)
 			CRTC_FB(crtc)->format->cpp[0] * 8,
 #else
 			CRTC_FB(crtc)->bits_per_pixel,
@@ -109,10 +107,10 @@ static void vbox_do_modeset(struct drm_crtc *crtc,
 		 0 : VBVA_SCREEN_F_BLANK;
 	flags |= vbox_crtc->disconnected ? VBVA_SCREEN_F_DISABLED : 0;
 	VBoxHGSMIProcessDisplayInfo(vbox->guest_pool, vbox_crtc->crtc_id,
-				    x_offset, y_offset, vbox_crtc->fb_offset +
-				    crtc->x * bpp / 8 + crtc->y * pitch,
-				    pitch, width, height,
-				    vbox_crtc->blanked ? 0 : bpp, flags);
+				   x_offset, y_offset, vbox_crtc->fb_offset +
+				   crtc->x * bpp / 8 + crtc->y * pitch,
+				   pitch, width, height,
+				   vbox_crtc->blanked ? 0 : bpp, flags);
 }
 
 static void vbox_crtc_dpms(struct drm_crtc *crtc, int mode)
@@ -176,16 +174,16 @@ static bool vbox_set_up_input_mapping(struct vbox_private *vbox)
 	if (single_framebuffer) {
 		list_for_each_entry(crtci, &vbox->dev->mode_config.crtc_list,
 				    head) {
-                        if (to_vbox_crtc(crtci)->crtc_id != 0)
-                                continue;
+			if (to_vbox_crtc(crtci)->crtc_id != 0)
+				continue;
 
-                        if (!CRTC_FB(crtci))
-                                break;
+			if (!CRTC_FB(crtci))
+				break;
 			vbox->single_framebuffer = true;
 			vbox->input_mapping_width = CRTC_FB(crtci)->width;
 			vbox->input_mapping_height = CRTC_FB(crtci)->height;
 			return old_single_framebuffer !=
-			    vbox->single_framebuffer;
+			       vbox->single_framebuffer;
 		}
 	}
 	/* Otherwise calculate the total span of all screens. */
@@ -209,7 +207,9 @@ static bool vbox_set_up_input_mapping(struct vbox_private *vbox)
 }
 
 static int vbox_crtc_set_base(struct drm_crtc *crtc,
-				 struct drm_framebuffer *old_fb, int x, int y)
+				 struct drm_framebuffer *old_fb,
+				 struct drm_framebuffer *new_fb,
+				 int x, int y)
 {
 	struct vbox_private *vbox = crtc->dev->dev_private;
 	struct vbox_crtc *vbox_crtc = to_vbox_crtc(crtc);
@@ -219,7 +219,7 @@ static int vbox_crtc_set_base(struct drm_crtc *crtc,
 	int ret;
 	u64 gpu_addr;
 
-	vbox_fb = to_vbox_framebuffer(CRTC_FB(crtc));
+	vbox_fb = to_vbox_framebuffer(new_fb);
 	obj = vbox_fb->obj;
 	bo = gem_to_vbox_bo(obj);
 
@@ -227,7 +227,7 @@ static int vbox_crtc_set_base(struct drm_crtc *crtc,
 	if (ret)
 		return ret;
 
-	ret = vbox_bo_pin(bo, TTM_PL_FLAG_VRAM, &gpu_addr);
+	ret = vbox_bo_pin(bo, VBOX_MEM_TYPE_VRAM, &gpu_addr);
 	vbox_bo_unreserve(bo);
 	if (ret)
 		return ret;
@@ -247,6 +247,9 @@ static int vbox_crtc_set_base(struct drm_crtc *crtc,
 		}
 	}
 
+	if (&vbox->fbdev->afb == vbox_fb)
+		vbox_fbdev_set_base(vbox, gpu_addr);
+
 	vbox_crtc->fb_offset = gpu_addr;
 	if (vbox_set_up_input_mapping(vbox)) {
 		struct drm_crtc *crtci;
@@ -266,17 +269,56 @@ static int vbox_crtc_mode_set(struct drm_crtc *crtc,
 			      int x, int y, struct drm_framebuffer *old_fb)
 {
 	struct vbox_private *vbox = crtc->dev->dev_private;
-	int rc = vbox_crtc_set_base(crtc, old_fb, x, y);
-	if (rc)
-		return rc;
+	int ret = vbox_crtc_set_base(crtc, old_fb, CRTC_FB(crtc), x, y);
+	if (ret)
+		return ret;
 	mutex_lock(&vbox->hw_mutex);
 	vbox_do_modeset(crtc, mode);
 	VBoxHGSMIUpdateInputMapping(vbox->guest_pool, 0, 0,
-				    vbox->input_mapping_width,
-				    vbox->input_mapping_height);
+				   vbox->input_mapping_width,
+				   vbox->input_mapping_height);
 	mutex_unlock(&vbox->hw_mutex);
 
-	return rc;
+	return ret;
+}
+
+static int vbox_crtc_page_flip(struct drm_crtc *crtc,
+			       struct drm_framebuffer *fb,
+#if RTLNX_VER_MIN(4,12,0) || RTLNX_RHEL_MAJ_PREREQ(7,5)
+			       struct drm_pending_vblank_event *event,
+			       uint32_t page_flip_flags,
+			       struct drm_modeset_acquire_ctx *ctx)
+#elif RTLNX_VER_MIN(3,12,0) || RTLNX_RHEL_MAJ_PREREQ(7,0)
+			       struct drm_pending_vblank_event *event,
+			       uint32_t page_flip_flags)
+#else
+			       struct drm_pending_vblank_event *event)
+#endif
+{
+	struct vbox_private *vbox = crtc->dev->dev_private;
+	struct drm_device *drm = vbox->dev;
+	unsigned long flags;
+	int rc;
+
+	rc = vbox_crtc_set_base(crtc, CRTC_FB(crtc), fb, 0, 0);
+	if (rc)
+		return rc;
+
+	vbox_do_modeset(crtc, &crtc->mode);
+	mutex_unlock(&vbox->hw_mutex);
+
+	spin_lock_irqsave(&drm->event_lock, flags);
+
+	if (event)
+#if RTLNX_VER_MIN(3,19,0) || RTLNX_RHEL_MAJ_PREREQ(7,2)
+		drm_crtc_send_vblank_event(crtc, event);
+#else
+		drm_send_vblank_event(drm, -1, event);
+#endif
+
+	spin_unlock_irqrestore(&drm->event_lock, flags);
+
+	return 0;
 }
 
 static void vbox_crtc_disable(struct drm_crtc *crtc)
@@ -316,6 +358,7 @@ static const struct drm_crtc_funcs vbox_crtc_funcs = {
 	.reset = vbox_crtc_reset,
 	.set_config = drm_crtc_helper_set_config,
 	/* .gamma_set = vbox_crtc_gamma_set, */
+	.page_flip = vbox_crtc_page_flip,
 	.destroy = vbox_crtc_destroy,
 };
 
@@ -342,7 +385,7 @@ static void vbox_encoder_destroy(struct drm_encoder *encoder)
 	kfree(encoder);
 }
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 13, 0) && !defined(RHEL_71)
+#if RTLNX_VER_MAX(3,13,0) && !RTLNX_RHEL_MAJ_PREREQ(7,1)
 static struct drm_encoder *drm_encoder_find(struct drm_device *dev, u32 id)
 {
 	struct drm_mode_object *mo;
@@ -355,26 +398,23 @@ static struct drm_encoder *drm_encoder_find(struct drm_device *dev, u32 id)
 static struct drm_encoder *vbox_best_single_encoder(struct drm_connector
 						    *connector)
 {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)
+#if RTLNX_VER_MIN(5,5,0) || RTLNX_RHEL_MIN(8,3)
         struct drm_encoder *encoder;
 
         /* There is only one encoder per connector */
         drm_connector_for_each_possible_encoder(connector, encoder)
             return encoder;
-#else /* KERNEL_VERSION < 5.5 */
+#else /* < 5.5 || RHEL < 8.3 */
 	int enc_id = connector->encoder_ids[0];
 
 	/* pick the encoder ids */
 	if (enc_id)
-# if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0) || \
-     (defined(CONFIG_SUSE_VERSION) && \
-         LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0)) || \
-     defined(RHEL_76)
+# if RTLNX_VER_MIN(4,15,0) || RTLNX_RHEL_MAJ_PREREQ(7,6) || (defined(CONFIG_SUSE_VERSION) && RTLNX_VER_MIN(4,12,0))
 		return drm_encoder_find(connector->dev, NULL, enc_id);
 # else
 		return drm_encoder_find(connector->dev, enc_id);
 # endif
-#endif /* KERNEL_VERSION < 5.5 */
+#endif /* < 5.5 || RHEL < 8.3 */
 	return NULL;
 }
 
@@ -425,11 +465,11 @@ static struct drm_encoder *vbox_encoder_init(struct drm_device *dev,
 		return NULL;
 
 	drm_encoder_init(dev, &vbox_encoder->base, &vbox_enc_funcs,
-			 DRM_MODE_ENCODER_DAC
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0) || defined(RHEL_73)
-			 , NULL
+#if RTLNX_VER_MIN(4,5,0) || RTLNX_RHEL_MAJ_PREREQ(7,3)
+			 DRM_MODE_ENCODER_DAC, NULL);
+#else
+			 DRM_MODE_ENCODER_DAC);
 #endif
-	    );
 	drm_encoder_helper_add(&vbox_encoder->base, &vbox_enc_helper_funcs);
 
 	vbox_encoder->base.possible_crtcs = 1 << i;
@@ -504,7 +544,7 @@ static void vbox_set_edid(struct drm_connector *connector, int width,
 	for (i = 0; i < EDID_SIZE - 1; ++i)
 		sum += edid[i];
 	edid[EDID_SIZE - 1] = (0x100 - (sum & 0xFF)) & 0xFF;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0) || defined(OPENSUSE_151) || defined(RHEL_77) || defined(RHEL_81)
+#if RTLNX_VER_MIN(4,19,0) || RTLNX_RHEL_MAJ_PREREQ(7,7) || RTLNX_RHEL_MAJ_PREREQ(8,1) || RTLNX_SUSE_MAJ_PREREQ(15,1) || RTLNX_SUSE_MAJ_PREREQ(12,5)
 	drm_connector_update_edid_property(connector, (struct edid *)edid);
 #else
 	drm_mode_connector_update_edid_property(connector, (struct edid *)edid);
@@ -534,7 +574,7 @@ static int vbox_get_modes(struct drm_connector *connector)
 	 * capability.
 	 */
 	VBoxHGSMIReportFlagsLocation(vbox->guest_pool, GUEST_HEAP_OFFSET(vbox) +
-				     HOST_FLAGS_OFFSET);
+				    HOST_FLAGS_OFFSET);
 	if (vbox_connector->vbox_crtc->crtc_id == 0)
 		vbox_report_caps(vbox);
 	if (!vbox->initial_mode_queried) {
@@ -562,7 +602,7 @@ static int vbox_get_modes(struct drm_connector *connector)
 	}
 	vbox_set_edid(connector, preferred_width, preferred_height);
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0) || defined(RHEL_72)
+#if RTLNX_VER_MIN(3,19,0) || RTLNX_RHEL_MAJ_PREREQ(7,2)
 	if (vbox_connector->vbox_crtc->x_hint != -1)
 		drm_object_property_set_value(&connector->base,
 			vbox->dev->mode_config.suggested_x_property,
@@ -583,7 +623,11 @@ static int vbox_get_modes(struct drm_connector *connector)
 	return num_modes;
 }
 
+#if RTLNX_VER_MAX(3,14,0) && !RTLNX_RHEL_MAJ_PREREQ(7,1)
 static int vbox_mode_valid(struct drm_connector *connector,
+#else
+static enum drm_mode_status vbox_mode_valid(struct drm_connector *connector,
+#endif
 			   struct drm_display_mode *mode)
 {
 	return MODE_OK;
@@ -591,10 +635,7 @@ static int vbox_mode_valid(struct drm_connector *connector,
 
 static void vbox_connector_destroy(struct drm_connector *connector)
 {
-	struct vbox_connector *vbox_connector = NULL;
-
-	vbox_connector = to_vbox_connector(connector);
-#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0) && !defined(RHEL_72)
+#if RTLNX_VER_MAX(3,17,0) && !RTLNX_RHEL_MAJ_PREREQ(7,2)
 	drm_sysfs_connector_remove(connector);
 #else
 	drm_connector_unregister(connector);
@@ -606,9 +647,8 @@ static void vbox_connector_destroy(struct drm_connector *connector)
 static enum drm_connector_status
 vbox_connector_detect(struct drm_connector *connector, bool force)
 {
-	struct vbox_connector *vbox_connector = NULL;
+	struct vbox_connector *vbox_connector;
 
-	(void)force;
 	vbox_connector = to_vbox_connector(connector);
 
 	return vbox_connector->mode_hint.disconnected ?
@@ -666,20 +706,20 @@ static int vbox_connector_init(struct drm_device *dev,
 	connector->interlace_allowed = 0;
 	connector->doublescan_allowed = 0;
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0) || defined(RHEL_72)
+#if RTLNX_VER_MIN(3,19,0) || RTLNX_RHEL_MAJ_PREREQ(7,2)
 	drm_mode_create_suggested_offset_properties(dev);
 	drm_object_attach_property(&connector->base,
 				   dev->mode_config.suggested_x_property, 0);
 	drm_object_attach_property(&connector->base,
 				   dev->mode_config.suggested_y_property, 0);
 #endif
-#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0) && !defined(RHEL_72)
+#if RTLNX_VER_MAX(3,17,0) && !RTLNX_RHEL_MAJ_PREREQ(7,2)
 	drm_sysfs_connector_add(connector);
 #else
 	drm_connector_register(connector);
 #endif
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0) || defined(OPENSUSE_151) || defined(RHEL_77) || defined(RHEL_81)
+#if RTLNX_VER_MIN(4,19,0) || RTLNX_RHEL_MAJ_PREREQ(7,7) || RTLNX_RHEL_MAJ_PREREQ(8,1) || RTLNX_SUSE_MAJ_PREREQ(15,1) || RTLNX_SUSE_MAJ_PREREQ(12,5)
 	drm_connector_attach_encoder(connector, encoder);
 #else
 	drm_mode_connector_attach_encoder(connector, encoder);
@@ -694,6 +734,7 @@ int vbox_mode_init(struct drm_device *dev)
 	struct drm_encoder *encoder;
 	struct vbox_crtc *vbox_crtc;
 	unsigned int i;
+	int ret;
 
 	/* vbox_cursor_init(dev); */
 	for (i = 0; i < vbox->num_crtcs; ++i) {
@@ -703,7 +744,9 @@ int vbox_mode_init(struct drm_device *dev)
 		encoder = vbox_encoder_init(dev, i);
 		if (!encoder)
 			return -ENOMEM;
-		vbox_connector_init(dev, vbox_crtc, encoder);
+		ret = vbox_connector_init(dev, vbox_crtc, encoder);
+		if (ret)
+			return ret;
 	}
 
 	return 0;
@@ -738,15 +781,15 @@ static int vbox_cursor_set2(struct drm_crtc *crtc, struct drm_file *file_priv,
 {
 	struct vbox_private *vbox = crtc->dev->dev_private;
 	struct vbox_crtc *vbox_crtc = to_vbox_crtc(crtc);
-	struct drm_gem_object *obj;
-	struct vbox_bo *bo;
-	int ret, rc;
 	struct ttm_bo_kmap_obj uobj_map;
-	u8 *src;
-	u8 *dst = NULL;
-	u32 caps = 0;
 	size_t data_size, mask_size;
+	struct drm_gem_object *obj;
+	u32 flags, caps = 0;
+	struct vbox_bo *bo;
 	bool src_isiomem;
+	u8 *dst = NULL;
+	u8 *src;
+	int ret;
 
 	if (!handle) {
 		bool cursor_enabled = false;
@@ -755,93 +798,98 @@ static int vbox_cursor_set2(struct drm_crtc *crtc, struct drm_file *file_priv,
 		/* Hide cursor. */
 		vbox_crtc->cursor_enabled = false;
 		list_for_each_entry(crtci, &vbox->dev->mode_config.crtc_list,
-				    head)
+				    head) {
 			if (to_vbox_crtc(crtci)->cursor_enabled)
 				cursor_enabled = true;
+			}
 
 		if (!cursor_enabled)
 			VBoxHGSMIUpdatePointerShape(vbox->guest_pool, 0, 0, 0,
-						    0, 0, NULL, 0);
+						   0, 0, NULL, 0);
 		return 0;
 	}
+
 	vbox_crtc->cursor_enabled = true;
+
 	if (width > VBOX_MAX_CURSOR_WIDTH || height > VBOX_MAX_CURSOR_HEIGHT ||
 	    width == 0 || height == 0)
 		return -EINVAL;
-	rc = VBoxQueryConfHGSMI(vbox->guest_pool,
+	ret = VBoxQueryConfHGSMI(vbox->guest_pool,
 				VBOX_VBVA_CONF32_CURSOR_CAPABILITIES, &caps);
-	ret = rc == VINF_SUCCESS ? 0 : rc == VERR_NO_MEMORY ? -ENOMEM : -EINVAL;
 	if (ret)
-		return ret;
+		return ret == VERR_NO_MEMORY ? -ENOMEM : -EINVAL;
 
-	if (!(caps & VBOX_VBVA_CURSOR_CAPABILITY_HARDWARE))
+	if (!(caps & VBOX_VBVA_CURSOR_CAPABILITY_HARDWARE)) {
 		/*
 		 * -EINVAL means cursor_set2() not supported, -EAGAIN means
 		 * retry at once.
 		 */
 		return -EBUSY;
+	}
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 7, 0) || defined(RHEL_74)
+#if RTLNX_VER_MIN(4,7,0) || RTLNX_RHEL_MAJ_PREREQ(7,4)
 	obj = drm_gem_object_lookup(file_priv, handle);
 #else
 	obj = drm_gem_object_lookup(crtc->dev, file_priv, handle);
 #endif
-	if (obj) {
-		bo = gem_to_vbox_bo(obj);
-		ret = vbox_bo_reserve(bo, false);
-		if (!ret) {
-			/*
-			 * The mask must be calculated based on the alpha
-			 * channel, one bit per ARGB word, and must be 32-bit
-			 * padded.
-			 */
-			mask_size = ((width + 7) / 8 * height + 3) & ~3;
-			data_size = width * height * 4 + mask_size;
-			vbox->cursor_hot_x = hot_x;
-			vbox->cursor_hot_y = hot_y;
-			vbox->cursor_width = width;
-			vbox->cursor_height = height;
-			vbox->cursor_data_size = data_size;
-			dst = vbox->cursor_data;
-			ret =
-			    ttm_bo_kmap(&bo->bo, 0, bo->bo.num_pages,
-					&uobj_map);
-			if (!ret) {
-				src =
-				    ttm_kmap_obj_virtual(&uobj_map,
-							 &src_isiomem);
-				if (!src_isiomem) {
-					u32 flags =
-					    VBOX_MOUSE_POINTER_VISIBLE |
-					    VBOX_MOUSE_POINTER_SHAPE |
-					    VBOX_MOUSE_POINTER_ALPHA;
-					copy_cursor_image(src, dst, width,
-							  height, mask_size);
-					rc = VBoxHGSMIUpdatePointerShape(
-						vbox->guest_pool, flags,
-						vbox->cursor_hot_x,
-						vbox->cursor_hot_y,
-						width, height, dst, data_size);
-					ret =
-					    rc == VINF_SUCCESS ? 0 : rc ==
-					    VERR_NO_MEMORY ? -ENOMEM : rc ==
-					    VERR_NOT_SUPPORTED ? -EBUSY :
-					    -EINVAL;
-				} else {
-					DRM_ERROR("src cursor bo should be in main memory\n");
-				}
-				ttm_bo_kunmap(&uobj_map);
-			} else {
-				vbox->cursor_data_size = 0;
-			}
-			vbox_bo_unreserve(bo);
-		}
-		drm_gem_object_put_unlocked(obj);
-	} else {
+	if (!obj) {
 		DRM_ERROR("Cannot find cursor object %x for crtc\n", handle);
-		ret = -ENOENT;
+		return -ENOENT;
 	}
 
+	bo = gem_to_vbox_bo(obj);
+	ret = vbox_bo_reserve(bo, false);
+	if (ret)
+		goto out_unref_obj;
+
+	/*
+	 * The mask must be calculated based on the alpha
+	 * channel, one bit per ARGB word, and must be 32-bit
+	 * padded.
+	 */
+	mask_size = ((width + 7) / 8 * height + 3) & ~3;
+	data_size = width * height * 4 + mask_size;
+	vbox->cursor_hot_x = hot_x;
+	vbox->cursor_hot_y = hot_y;
+	vbox->cursor_width = width;
+	vbox->cursor_height = height;
+	vbox->cursor_data_size = data_size;
+	dst = vbox->cursor_data;
+
+	ret = ttm_bo_kmap(&bo->bo, 0, bo->bo.num_pages, &uobj_map);
+	if (ret) {
+		vbox->cursor_data_size = 0;
+		goto out_unreserve_bo;
+	}
+
+	src = ttm_kmap_obj_virtual(&uobj_map, &src_isiomem);
+	if (src_isiomem) {
+		DRM_ERROR("src cursor bo not in main memory\n");
+		ret = -EIO;
+		goto out_unmap_bo;
+	}
+
+	copy_cursor_image(src, dst, width, height, mask_size);
+
+	flags = VBOX_MOUSE_POINTER_VISIBLE | VBOX_MOUSE_POINTER_SHAPE |
+		VBOX_MOUSE_POINTER_ALPHA;
+	ret = VBoxHGSMIUpdatePointerShape(vbox->guest_pool, flags,
+					 vbox->cursor_hot_x, vbox->cursor_hot_y,
+					 width, height, dst, data_size);
+	ret = ret == VINF_SUCCESS ? 0 : ret == VERR_NO_MEMORY ? -ENOMEM :
+		ret == VERR_NOT_SUPPORTED ? -EBUSY : -EINVAL;
+
+out_unmap_bo:
+	ttm_bo_kunmap(&uobj_map);
+out_unreserve_bo:
+	vbox_bo_unreserve(bo);
+out_unref_obj:
+#if RTLNX_VER_MIN(5,9,0)
+	drm_gem_object_put(obj);
+#else
+	drm_gem_object_put_unlocked(obj);
+#endif
+
 	return ret;
 }
 
@@ -852,7 +900,7 @@ static int vbox_cursor_move(struct drm_crtc *crtc, int x, int y)
 	    vbox->single_framebuffer ? crtc->x : to_vbox_crtc(crtc)->x_hint;
 	s32 crtc_y =
 	    vbox->single_framebuffer ? crtc->y : to_vbox_crtc(crtc)->y_hint;
-	int rc;
+	int ret;
 
 	x += vbox->cursor_hot_x;
 	y += vbox->cursor_hot_y;
@@ -861,8 +909,8 @@ static int vbox_cursor_move(struct drm_crtc *crtc, int x, int y)
 		y + crtc_y >= vbox->input_mapping_width ||
 		vbox->cursor_data_size == 0)
 		return 0;
-	rc = VBoxHGSMICursorPosition(vbox->guest_pool, true, x + crtc_x,
+	ret = VBoxHGSMICursorPosition(vbox->guest_pool, true, x + crtc_x,
 					 y + crtc_y, NULL, NULL);
-	return rc == VINF_SUCCESS ? 0 : rc == VERR_NO_MEMORY ? -ENOMEM : rc ==
+	return ret == VINF_SUCCESS ? 0 : ret == VERR_NO_MEMORY ? -ENOMEM : ret ==
 		VERR_NOT_SUPPORTED ? -EBUSY : -EINVAL;
 }
